// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import elliptic from 'elliptic';
import utils from './utils';
import BigInt from 'big-integer';
import BN from 'bn.js';
import Hash from './hash';
import EcCurve from './ecCurve';

/**
 * @typedef {string | number | BigInt.BigInteger| BN } BNInput
 */

export default class EcPointBase {
    /** @returns {EcCurve} */
    get _curve() { throw new Error("Method not implemented."); }
    
    get x() { return BigInt.fromArray(this._value.getX().toArray('be'), 256, false); }

    get y() { return BigInt.fromArray(this._value.getY().toArray('be'), 256, false); }

    get isNegative() { return !this.isInfinity && !this._value.eq(this._curve.point(this._value.getX())); }

    get isInfinity() { return this._value.isInfinity(); }

    get isValid() { return this._value.validate(); }

    get axisSize() { return Math.ceil(this._curve.p.bitLength() / 8); }

    /**
     * @param {BNInput|elliptic.curve.base.BasePoint} x
     * @param {BNInput} y
     */
    constructor(x = null, y = null) {
        if (x instanceof elliptic.curve.base.BasePoint)
            this._value = x != null ? x : this._curve.infinity;
        else
            this._value = this._curve.point(getBN(x), getBN(y));
    }

    /**
     * @private
     * @param {elliptic.curve.base.BasePoint} [point]
     */
    _create(point) {
        /** @type {EcPointBase} */
        var pnt = Object.create(Object.getPrototypeOf(this));
        pnt._value = point;
        return pnt;
    }

    /** @param {EcPointBase} point */
    plus(point) {
        return this.add(point)
    }

    /** @param {EcPointBase} point */
    add(point) {
        return this._create(this._value.add(point._value))
    }

    /** @param {BNInput} num */
    mul(num) {
        return this._create(this._value.mul(getBN(num)));
    }

    /** @param {BNInput} num */
    multiply(num) {
        return this.times(num);
    }    

    /** @param {BNInput} num */
    times(num) {
        return this._create(this._value.mul(getBN(num)));
    }

    negate() {
        return this._create(this._value.neg());
    }

    /** @param {EcPointBase} point */
    subtract(point) {
        return this.minus(point);
    }

    /** @param {EcPointBase} point */
    minus(point) {
        return this.add(point.negate());
    }

    /** @param {EcPointBase} point */
    eq(point) {
        return this._value.eq(point._value);
    }

    values() {
        return [this.x, this.y];
    }

    /** @returns {Uint8Array} */
    toArray(compact = false) {
        if (!compact)
            return new Uint8Array(utils.padLeft(this._value.getX().toArray('be'), 32)
                .concat(utils.padLeft(this._value.getY().toArray('be'), 32)));

        var val = this._value.getX().shln(1).or(new BN(this.isNegative ? 1 : 0));
        return new Uint8Array(utils.padLeft(val.toArray('be'), this.axisSize));
    }

    /**
     * @protected
     * @param {EcCurve} curve
     * @param {Uint8Array} data
     */
    static _from(curve, data, compact = false) {
        var point = create(EcPointBase, curve.point());
        var size = Math.ceil(curve.p.bitLength() / 8);
        
        if (!compact) {
            point._value = curve.point(new BN(data.slice(0, size), 'be'), new BN(data.slice(size), 'be'));
            return point;
        }

        var val = new BN(data);
        var isNeg = val.testn(0);
        point._value = curve.point(val.shrn(1));

        return isNeg ? point.negate() : point;
    }

    /**
     * @protected
     * @param {EcCurve} curve
     * @param {string} message
     * @returns {EcPointBase}
     */
    static _fromString(curve, message) {
        var [p, order, one] = [curve.p, curve.order, new BN(1)];
        var x = new BN(Array.from(Hash.shaBuffer(message)), 'be').mod(p);
        var point = create(EcPointBase, curve.point());

        while (point.isInfinity) {
            try {
                var pnt = curve.point(x);
                if (!order || pnt.mul(order).isInfinity())
                    point._value = pnt;
            }
            catch { }
            finally { x.iadd(one); }
        }

        return point;
    }

    inspect() {
        return `${this.constructor.name} {x: ${this.x.toString(16)}, y: ${this.y.toString(16)}}`;
    }
}

/** @param {BNInput} number */
function getBN(number) {
    if (number instanceof BigInt)
        return utils.ToBn(number);
    else if (number instanceof BN || typeof number == "string" || typeof number == "number")
        return new BN(number);
}

/**
 * @private
 * @param {typeof EcPointBase} [type]
 * @param {elliptic.curve.base.BasePoint} [point]
 * @returns {EcPointBase}
 */
function create(type, point) {
    /** @type {EcPointBase} */
    var pnt = Object.create(type.prototype);
    pnt._value = point;
    return pnt;
}
