// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import utils, { constants } from './utils.js';

export default class SecretShare {
    /**
     * @param {bigInt.BigInteger} xi
     * @param {bigInt.BigInteger[]} xs
     * @param {bigInt.BigInteger} m
     * @returns {bigInt.BigInteger}
     */
    static getLi(xi, xs, m) {
        return basisPoly(xi, xs, m)
    }

    static interpolatePartial(xs, ys, p, q) {
        return xs.map((xi, i) => ys[i].modPow(basisPoly(xi, xs, q), p))
            .reduce((secret, num) => secret.times(num).mod(p));
    }

    static interpolate(xs, ys, m) {
        return xs.map((xi, i) => basisPoly(xi, xs, m).times(ys[i]).mod(m))
            .reduce((secret, num) => secret.plus(num).mod(m));
    }

    /**
     * @param {bigInt.BigInteger} secret
     * @param {number} number
     * @param {number} threshold
     * @param {bigInt.BigInteger} m
     */
    static share(secret, number, threshold, m) {
        const xs = randoms(number, m);
        return SecretShare.shareFromIds(secret, xs, threshold, m);
    }

    /**
     * @param {bigInt.BigInteger} secret
     * @param {bigInt.BigInteger[]} ids
     * @param {number} threshold
     * @param {bigInt.BigInteger} m
     */
    static shareFromIds(secret, ids, threshold, m) {
        const coeffs = [secret].concat(randoms(threshold - 1, m));
        return [ids, ids.map(x => evalPoly(coeffs, x, m))];
    }
    
    static interpolateText(shares, m) {
        try {
            return SecretShare.interpolateBuffer(shares, m).toString('utf8');
        } catch { return ""; }
    }

    //TODO: Verify the input
    static interpolateBuffer(shares, m) {
        const size = m.bitLength().value;
        const bytes = Math.ceil(size / 8);

        shares = shares.map(shr => Buffer.from(shr, 'base64'));
        const xs = shares.map(shr => utils.decode(shr.slice(0, bytes), size)[0]);
        const chunks = shares.map(shr => utils.decode(shr.slice(bytes), size));

        let numbers = [];
        for (let i = 0; i < chunks[0].length; i++) {
            const ys = chunks.map(ch => ch[i]);
            numbers.push(SecretShare.interpolate(xs, ys, m));
        }
        
        return Buffer.from(utils.encodeByteArray(numbers, size));
    }

    static shareText(data, number, threshold, m) {
        return SecretShare.shareBuffer(data, number, threshold, m).map(sh => sh.toString('base64'));
    }

    static shareBuffer(data, number, threshold, m) {
        const size = m.bitLength().value;
        const chunks = utils.decode(data, size, 'utf8');
        
        const shares = randoms(number, m).map(xi => ({xi, ys: []}));
        for (let i = 0; i < chunks.length; i++) {
            const coeffs = [chunks[i]].concat(randoms(threshold - 1, m));
            for (let j = 0; j < shares.length; j++) {
                shares[j].ys.push(evalPoly(coeffs, shares[j].xi, m));
            }
        }

        return shares.map(sh => [utils.encodeByteArray([sh.xi], size, true), utils.encodeByteArray(sh.ys, size, true)])
            .map(sh => Buffer.from(Array.prototype.concat.apply([], sh)));
    }
}

function basisPoly(xi, xs, m) {
    var li = xs.filter(xj => !xj.equals(xi))
        .map(xj => xj.minus(xi).modInv(m).times(xj).mod(m))
        .reduce((li, num) => li.times(num).mod(m));
    return li.sign ? m.plus(li) : li; //library does not support unsigned mod (╯°□°)╯
}

function randoms(count, m) {
    var numbers = [];
    while (numbers.length < count) {
        var rdm = utils.random(constants.one, m.minus(constants.one));
        if (numbers.every(num => !rdm.equals(num)))
            numbers.push(rdm);
    }
    
    return numbers;
}

function evalPoly(coeffs, x, m) {
    let y = coeffs[coeffs.length - 1];
    for (let i = coeffs.length - 2; i >= 0; i--) {
        y = y.times(x).mod(m).plus(coeffs[i]).mod(m);
    }
    return y;
}

