// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import TObliviousClient from './tobliviousClient';
import SecretShare from './secretShare';
import BigInt from 'big-integer';
import ElGamalKey from './elGamalKey';
import utils from './utils';
import C25519Point from './c25519Point';

export default class TObliviousFlow {
    /**
     * @param {string[]} urls
     * @param {string} user
     */
    constructor(urls, user) {
        this.clients = urls.map(url => new TObliviousClient(url, user));
        this.user = user;
    }

    /**
     * @param {bigInt.BigInteger} key
     * @param {number} threshold
     */
    async signIn(key, threshold) {
        try {
            var ids = this.clients.map(c => c.id);
            var [, kis] = SecretShare.shareFromIds(key, ids, threshold, C25519Point.n);

            await Promise.all(this.clients.map((cli, i) => cli.SetSecret(kis[i])));
        } catch (err) {
            return Promise.reject(err);
        }
    }

    /** @param {string} password */
    async logIn(password) {
        try {
            var r = random();
            var n = C25519Point.n;
            var g = C25519Point.fromString(password);
            var gR = g.multiply(r);

            var gRKis = await Promise.all(this.clients.map(cli => cli.GetSecret(gR)));
            var ids = this.clients.map(c => c.id);
            var lis = ids.map(id => SecretShare.getLi(id, ids, n));

            var gRK = gRKis.map((ki, i) => ki.times(lis[i])).reduce((rki, sum) => rki.add(sum));
            var gK = gRK.times(r.modInv(n));

            var [s, p] = ElGamalKey.seed(gK.toArray());

            return { priv: s, pub: p };
        } catch (err) {
            return Promise.reject(err);
        }
    }
}

function random() {
    return utils.random(BigInt.one, C25519Point.n.subtract(BigInt.one));
}